"""
Constants for pynws
"""
import os

from .version import __version__

file_dir = os.path.join(os.path.dirname(__file__), '..')

version = __version__

API_URL = 'https://api.weather.gov/'
API_STATIONS = 'points/{},{}/stations'
API_OBSERVATION = 'stations/{}/observations/'
API_ACCEPT = 'application/geo+json'
API_USER = 'pynws {}'
API_FORECAST = 'points/{},{}/forecast'
API_FORECAST_HOURLY = '/points/{},{}/forecast/hourly'
API_GRID_FORECAST = 'gridpoints/{}/{},{}/forecast'
API_GRID_FORECAST_HOURLY = 'gridpoints/{}/{},{}/forecast/hourly'
API_POINT = 'points/{},{}'
API_ALERTS_ZONE = 'alerts/active/zone/{}'

DEFAULT_USERID = 'CODEemail@address'

API_WEATHER_CODE = {
    "skc": "Fair/clear",
    "few": "A few clouds",
    "sct": "Partly cloudy",
    "bkn": "Mostly cloudy",
    "ovc": "Overcast",
    "wind_skc": "Fair/clear and windy",
    "wind_few": "A few clouds and windy",
    "wind_sct": "Partly cloudy and windy",
    "wind_bkn": "Mostly cloudy and windy",
    "wind_ovc": "Overcast and windy",
    "snow": "Snow",
    "rain_snow": "Rain/snow",
    "rain_sleet": "Rain/sleet",
    "snow_sleet": "Snow/sleet",
    "fzra": "Freezing rain",
    "rain_fzra": "Rain/freezing rain",
    "snow_fzra": "Freezing rain/snow",
    "sleet": "Sleet",
    "rain": "Rain",
    "rain_showers": "Rain showers (high cloud cover)",
    "rain_showers_hi": "Rain showers (low cloud cover)",
    "tsra": "Thunderstorm (high cloud cover)",
    "tsra_sct": "Thunderstorm (medium cloud cover)",
    "tsra_hi": "Thunderstorm (low cloud cover)",
    "tornado": "Tornado",
    "hurricane": "Hurricane conditions",
    "tropical_storm": "Tropical storm conditions",
    "dust": "Dust",
    "smoke": "Smoke",
    "haze": "Haze",
    "hot": "Hot",
    "cold": "Cold",
    "blizzard": "Blizzard",
    "fog": "Fog/mist"
    }



"""Support for NWS weather service."""
from statistics import mean

from metar import Metar

from .nws import Nws
from .const import API_WEATHER_CODE

WIND_DIRECTIONS = ['N', 'NNE', 'NE', 'ENE',
                   'E', 'ESE', 'SE', 'SSE',
                   'S', 'SSW', 'SW', 'WSW',
                   'W', 'WNW', 'NW', 'NNW']

WIND = {name: idx * 360 / 16 for idx, name in enumerate(WIND_DIRECTIONS)}

OBSERVATIONS = {
    'temperature': ['temp', 'C', None],
    'barometricPressure': None,
    'seaLevelPressure': ['press', 'HPA', 100],
    'relativeHumidity': None,
    'windSpeed': ['wind_speed', 'MPS', None],
    'windDirection': ['wind_dir', None, None],
    'visibility': ['vis', 'M', None],
    'elevation': None,
    'textDescription': None,
    'dewpoint': None,
    'windGust': None,
    'station': None,
    'timestamp': None,
    'icon': None,
    'maxTemperatureLast24Hours': None,
    'minTemperatureLast24Hours': None,
    'precipitationLastHour': None,
    'precipitationLast3Hours': None,
    'precipitationLast6Hours': None,
    'windChill': None,
    'heatIndex': None,
}

def convert_weather(weather):
    """Convert short code to readable name."""
    return [(API_WEATHER_CODE.get(w[0], w[0]), w[1]) for w in weather]


def parse_icon(icon):
    """
    Parse icon url to NWS weather codes.

    Example:
    https://api.weather.gov/icons/land/day/skc/tsra,40?size=medium

    Example return:
    ('day', (('skc', None), ('tsra', 40),))
    """
    icon_list = icon.split('/')
    time = icon_list[5]
    weather = [i.split('?')[0] for i in icon_list[6:]]
    code = [w.split(',')[0] for w in weather]
    chance = [int(w.split(',')[1]) if len(w.split(',')) == 2 else None
              for w in weather]
    return time, tuple(zip(code, chance))


class SimpleNWS:
    """
    NWS simplified data.

    Uses normal api first.  If value is None, use metar info.
    """
    def __init__(self, lat, lon, api_key, session):
        """Set up simplified NWS class."""
        self.lat = lat
        self.lon = lon
        self.api_key = api_key
        self.session = session
        self.nws = Nws(session, latlon=(float(lat), float(lon)),
                       userid=api_key)

        self._observation = None
        self._metar_obs = None
        self.station = None
        self.stations = None
        self._forecast = None
        self._forecast_hourly = None
        self._alerts_zone = None

    async def set_station(self, station=None):
        """
        Set station or retrieve station list.
        If no station is supplied, the first retrieved value is set.
        """
        if station:
            self.nws.station = station
            self.station = station
            if not self.stations:
                self.stations = [self.station]
        else:
            self.stations = await self.nws.stations()
            self.nws.station = self.stations[0]
            self.station = self.stations[0]

    @staticmethod
    def extract_metar(obs):
        """Return parsed metar if available."""
        metar_msg = obs.get('rawMessage')
        if metar_msg:
            try:
                metar_obs = Metar.Metar(metar_msg)
            except Metar.ParserError:
                metar_obs = None
        else:
            metar_obs = None
        return metar_obs

    async def update_observation(self, limit=0, start_time=None):
        """Update observation."""
        obs = await self.nws.observations(limit, start_time=start_time)
        if obs is None:
            return None
        self._observation = obs
        self._metar_obs = [self.extract_metar(iobs) for iobs in self._observation]

    async def update_forecast(self):
        """Update forecast."""
        self._forecast = await self.nws.grid_forecast()

    async def update_forecast_hourly(self):
        """Update forecast."""
        self._forecast_hourly = await self.nws.grid_forecast_hourly()

    async def update_alerts_zone(self):
        """Update alerts zone."""
        self._alerts_zone = await self.nws.alerts_zone()

    @staticmethod
    def extract_observation_value(observation, value):
        """Returns observation or observation value."""
        obs_value = observation.get(value)
        if obs_value is None:
            return None
        if isinstance(observation[value], dict):
            obs_sub_value = observation[value].get('value')
            if obs_sub_value is None:
                return None
            return float(obs_sub_value)
        return observation[value]

    @property
    def observation(self):
        """Observation dict"""

        if self._observation is None:
            return None

        data = {}
        for obs, met in OBSERVATIONS.items():
            obs_list = [self.extract_observation_value(o, obs) for o in self._observation]
            data[obs] = next(iter([o for o in obs_list if o]), None)
            if data[obs] is None and (met is not None
                                      and self._metar_obs[0] is not None):
                met_prop = getattr(self._metar_obs[0], met[0])
                if met_prop:
                    if met[1]:
                        data[obs] = met_prop.value(units=met[1])
                    else:
                        data[obs] = met_prop.value()
                    if met[2] is not None:
                        data[obs] = data[obs] * met[2]
        if data.get('icon'):
            time, weather = parse_icon(data['icon'])
            data['iconTime'] = time
            data['iconWeather'] = convert_weather(weather)
        else:
            data['iconTime'] = None
            data['iconWeather'] = None
        return data

    @property
    def forecast(self):
        """Return forecast."""
        return self._convert_forecast(self._forecast)

    @property
    def forecast_hourly(self):
        """Return forecast hourly."""
        return self._convert_forecast(self._forecast_hourly)

    @staticmethod
    def _convert_forecast(input_forecast):
        """Converts forecast to common dict."""
        if not input_forecast:
            return []
        forecast = []
        for forecast_entry in input_forecast:
            # get weather
            if forecast_entry.get('temperature'):
                forecast_entry['temperature'] = int(forecast_entry['temperature'])

            if forecast_entry.get('icon'):
                time, weather = parse_icon(forecast_entry['icon'])
                weather = convert_weather(weather)
            else:
                time, weather = (None, None)
            forecast_entry['iconTime'] = time
            forecast_entry['iconWeather'] = weather
            if forecast_entry.get('windDirection'):
                forecast_entry['windBearing'] = \
                    WIND[forecast_entry['windDirection']]
            else:
                forecast_entry['windBearing'] = None

            # wind speed reported as '7 mph' or '7 to 10 mph'
            # if range, take average
            if forecast_entry.get('windSpeed'):
                wind_speed = forecast_entry['windSpeed'].split(' ')[0::2]
                wind_speed_avg = mean(int(w) for w in wind_speed)
                forecast_entry['windSpeedAvg'] = wind_speed_avg
            else:
                forecast_entry['windSpeedAvg'] = None

            forecast.append(forecast_entry)
        return forecast

    @property
    def alerts_zone(self):
        """Return alerts as a list of dict."""
        return self._alerts_zone



"""pynws module."""
from datetime import timedelta, datetime, timezone

import pynws.urls
from pynws.const import API_ACCEPT, API_USER, DEFAULT_USERID

class NwsError(Exception):
    """Error in Nws Class"""
    def __init__(self, message):
        super().__init__(message)
        self.message = message

class Nws:
    """Nws object for simple use"""
    def __init__(self, session, latlon=None, station=None, userid=DEFAULT_USERID):
        self.session = session
        self.latlon = latlon
        self.station = station
        self.userid = userid

        self.wfo = None
        self.x = None
        self.y = None

        self.zone = None

    async def stations(self):
        """Returns station list"""
        if self.latlon is None:
            raise NwsError("Need to set lattitude and longitude")
        return await stations(*self.latlon, self.session, self.userid)

    async def observations(self, limit=0, start_time=None):
        """Returns observation list"""
        if self.station is None:
            raise NwsError("Need to set station")
        return await observations(self.station, self.session, self.userid, limit, start_time)


    async def get_pointdata(self):
        """Saves griddata from latlon."""
        data = await get_pointdata(*self.latlon, self.session, self.userid)

        properties = data.get('properties')
        if properties:
            self.wfo = properties.get('cwa')
            self.x = properties.get('gridX')
            self.y = properties.get('gridY')
            self.zone = properties.get('forecastZone').split("/")[-1]

    async def grid_forecast(self):
        """Return forecast from grid."""
        if self.wfo is None:
            await self.get_pointdata()
        raw_forecast = await grid_forecast_raw(self.wfo, self.x, self.y, self.session, self.userid)
        return raw_forecast['properties']['periods']

    async def grid_forecast_hourly(self):
        """Return hourly forecast from grid."""
        if self.wfo is None:
            await self.get_pointdata()
        raw_forecast = await grid_forecast_hourly_raw(self.wfo, self.x, self.y,
                                                      self.session, self.userid)
        return raw_forecast['properties']['periods']

    async def forecast(self):
        """Returns forecast list"""
        if self.latlon is None:
            raise NwsError("Need to set lattitude and longitude")
        return await forecast(*self.latlon, self.session, self.userid)

    async def forecast_hourly(self):
        """returns hourly forecast list"""
        if self.latlon is None:
            raise NwsError("Need to set lattitude and longitude")
        return await forecast_hourly(*self.latlon, self.session, self.userid)

    async def alerts_zone(self):
        """Returns alerts dict."""
        if self.zone is None:
            await self.get_pointdata()
        alerts = await alerts_zone_raw(self.zone, self.session, self.userid)
        return [alert['properties'] for alert in alerts['features']]

def get_header(userid):
    """Get header.

    NWS recommends including an email in userid.
    """
    return {'accept': API_ACCEPT,
            'User-Agent': API_USER.format(userid)}


async def get_obs_from_stn(station, websession, userid, limit=5, start_time=None):
    """Get observation response from station"""
    params = {}
    if limit > 0:
        params['limit'] = limit

    if start_time:
        if not isinstance(start_time, timedelta):
            raise ValueError
        now = datetime.now(timezone.utc)
        request_time = now - start_time
        params['start'] = request_time.isoformat(timespec='seconds')

    url = pynws.urls.obs_url(station)
    header = get_header(userid)
    async with websession.get(url, headers=header, params=params) as res:
        res.raise_for_status()
        obs = await res.json()
    return obs


async def observations(station, websession, userid, limit=5, start_time=None):
    """Observations from station"""
    res = await get_obs_from_stn(station, websession, userid, limit, start_time)
    return [o['properties'] for o in res['features']]


async def get_stn_from_pnt(lat, lon, websession, userid):
    """Get list of stations for lat/lon"""

    url = pynws.urls.stn_url(lat, lon)
    header = get_header(userid)
    async with websession.get(url, headers=header) as res:
        res.raise_for_status()
        jres = await res.json()
    return jres


async def stations(lat, lon, websession, userid):
    """Returns list of stations for a point."""
    res = await get_stn_from_pnt(lat, lon, websession, userid)
    return [s['properties']['stationIdentifier']
            for s in res['features']]


async def get_forc_from_pnt(lat, lon, websession, userid):
    """update forecast"""

    url = pynws.urls.forc_url(lat, lon)
    header = get_header(userid)
    async with websession.get(url, headers=header) as res:
        res.raise_for_status()
        jres = await res.json()
    return jres


async def forecast(lat, lon, websession, userid):
    """Returns forecast as list """
    res = await get_forc_from_pnt(lat, lon, websession, userid)
    return res['properties']['periods']


async def get_hour_forc_from_pnt(lat, lon, websession, userid):
    """update forecast"""

    url = pynws.urls.hour_forc_url(lat, lon)
    header = get_header(userid)
    async with websession.get(url, headers=header) as res:
        res.raise_for_status()
        jres = await res.json()
    return jres


async def forecast_hourly(lat, lon, websession, userid):
    """Returns hourly forecast as list """
    res = await get_hour_forc_from_pnt(lat, lon, websession, userid)
    return res['properties']['periods']


async def get_pointdata(lat, lon, websession, userid):
    """Return griddata response."""

    url = pynws.urls.point_url(lat, lon)
    header = get_header(userid)
    async with websession.get(url, headers=header) as res:
        res.raise_for_status()
        jres = await res.json()
    return jres

async def grid_forecast_raw(wfo, x, y, websession, userid):
    """Return griddata response."""

    url = pynws.urls.grid_forecast_url(wfo, x, y)
    header = get_header(userid)
    async with websession.get(url, headers=header) as res:
        res.raise_for_status()
        jres = await res.json()
    return jres


async def grid_forecast_hourly_raw(wfo, x, y, websession, userid):
    """Return griddata response."""

    url = pynws.urls.grid_forecast_hourly_url(wfo, x, y)
    header = get_header(userid)
    async with websession.get(url, headers=header) as res:
        res.raise_for_status()
        jres = await res.json()
    return jres


async def alerts_zone_raw(zone, websession, userid):
    """Return griddata response."""

    url = pynws.urls.alerts_zone_url(zone)
    header = get_header(userid)
    async with websession.get(url, headers=header) as res:
        res.raise_for_status()
        jres = await res.json()
    return jres



"""
url formattera
"""

from pynws.const import (API_URL, API_ALERTS_ZONE, API_STATIONS, API_OBSERVATION, API_FORECAST,
                         API_FORECAST_HOURLY, API_GRID_FORECAST, API_GRID_FORECAST_HOURLY,
                         API_POINT)

def obs_url(station):
    """Formats observation url."""
    return API_URL + API_OBSERVATION.format(station)

def stn_url(lat, lon):
    """formats station url"""
    return API_URL + API_STATIONS.format(str(lat), str(lon))

def forc_url(lat, lon):
    """Formats forecast url"""
    return API_URL + API_FORECAST.format(lat, lon)

def hour_forc_url(lat, lon):
    """Formats forecast url"""
    return API_URL + API_FORECAST_HOURLY.format(lat, lon)

def grid_forecast_url(wfo, x, y):
    """gridpoint forecast."""
    return API_URL + API_GRID_FORECAST.format(wfo, x, y)

def grid_forecast_hourly_url(wfo, x, y):
    """gridpoint forecast hpurly."""
    return API_URL + API_GRID_FORECAST_HOURLY.format(wfo, x, y)

def point_url(lat, lon):
    """ Formats point metadata url."""
    return API_URL + API_POINT.format(lat, lon)

def alerts_zone_url(zone):
    """Formats alerts url with zone."""
    return API_URL + API_ALERTS_ZONE.format(zone)



import os
from setuptools import setup

exec(open('pynws/version.py').read())

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='pynws',
    version=__version__,
    license='MIT License',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/MatthewFlamm/pynws',
    author='Matthew Flamm',
    author_email='matthewflamm0@gmail.com',
    description='Python library to retrieve observations and forecasts from NWS/NOAA',
    packages=['pynws'],
    include_package_data=True,
    install_requires=[
        'aiohttp',
        'metar',
    ],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7"]
)



"""
This is a module for querying weather data from the NWS/NOAA
asynchronously and organizing the data in an easier to use manner
"""

from .simple_nws import SimpleNWS
from .nws import *
from .const import version
